@@ -10,6 +10,11 @@
  * and Ecole Normale Superieure, 45 rue d'Ulm, 75230 Paris, France
  */
 
+#undef TYPE
+#define TYPE	UNION
+static
+#include "has_single_reference_templ.c"
+
 __isl_give UNION *FN(UNION,cow)(__isl_take UNION *u);
 
 isl_ctx *FN(UNION,get_ctx)(__isl_keep UNION *u)
@@ -17,24 +22,33 @@ isl_ctx *FN(UNION,get_ctx)(__isl_keep UNION *u)
 	return u ? u->space->ctx : NULL;
 }
 
-__isl_give isl_space *FN(UNION,get_space)(__isl_keep UNION *u)
+/* Return the space of "u".
+ */
+static __isl_keep isl_space *FN(UNION,peek_space)(__isl_keep UNION *u)
 {
 	if (!u)
 		return NULL;
-	return isl_space_copy(u->space);
+	return u->space;
+}
+
+/* Return a copy of the space of "u".
+ */
+__isl_give isl_space *FN(UNION,get_space)(__isl_keep UNION *u)
+{
+	return isl_space_copy(FN(UNION,peek_space)(u));
 }
 
 /* Return the number of parameters of "u", where "type"
  * is required to be set to isl_dim_param.
  */
-unsigned FN(UNION,dim)(__isl_keep UNION *u, enum isl_dim_type type)
+isl_size FN(UNION,dim)(__isl_keep UNION *u, enum isl_dim_type type)
 {
 	if (!u)
-		return 0;
+		return isl_size_error;
 
 	if (type != isl_dim_param)
 		isl_die(FN(UNION,get_ctx)(u), isl_error_invalid,
-			"can only reference parameters", return 0);
+			"can only reference parameters", return isl_size_error);
 
 	return isl_space_dim(u->space, type);
 }
@@ -52,19 +66,19 @@ int FN(UNION,find_dim_by_name)(__isl_keep UNION *u, enum isl_dim_type type,
 }
 
 #ifdef HAS_TYPE
-static __isl_give UNION *FN(UNION,alloc)(__isl_take isl_space *dim,
+static __isl_give UNION *FN(UNION,alloc)(__isl_take isl_space *space,
 	enum isl_fold type, int size)
 #else
-static __isl_give UNION *FN(UNION,alloc)(__isl_take isl_space *dim, int size)
+static __isl_give UNION *FN(UNION,alloc)(__isl_take isl_space *space, int size)
 #endif
 {
 	UNION *u;
 
-	dim = isl_space_params(dim);
-	if (!dim)
+	space = isl_space_params(space);
+	if (!space)
 		return NULL;
 
-	u = isl_calloc_type(dim->ctx, UNION);
+	u = isl_calloc_type(space->ctx, UNION);
 	if (!u)
 		goto error;
 
@@ -72,13 +86,13 @@ static __isl_give UNION *FN(UNION,alloc)(__isl_take isl_space *dim, int size)
 #ifdef HAS_TYPE
 	u->type = type;
 #endif
-	u->space = dim;
-	if (isl_hash_table_init(dim->ctx, &u->table, size) < 0)
+	u->space = space;
+	if (isl_hash_table_init(space->ctx, &u->table, size) < 0)
 		return FN(UNION,free)(u);
 
 	return u;
 error:
-	isl_space_free(dim);
+	isl_space_free(space);
 	return NULL;
 }
 
@@ -108,21 +122,12 @@ __isl_give UNION *FN(UNION,copy)(__isl_keep UNION *u)
  * Return the ZERO element if "u" does not contain any element
  * living in "space".
  */
-__isl_give PART *FN(FN(UNION,extract),PARTS)(__isl_keep UNION *u,
+__isl_give PART *FN(FN(UNION,extract),BASE)(__isl_keep UNION *u,
 	__isl_take isl_space *space)
 {
 	struct isl_hash_table_entry *entry;
 
-	if (!u || !space)
-		goto error;
-	if (!isl_space_match(u->space, isl_dim_param, space, isl_dim_param)) {
-		space = isl_space_drop_dims(space, isl_dim_param,
-					0, isl_space_dim(space, isl_dim_param));
-		space = isl_space_align_params(space,
-					FN(UNION,get_space)(u));
-		if (!space)
-			goto error;
-	}
+	space = isl_space_replace_params(space, FN(UNION,peek_space)(u));
 
 	entry = FN(UNION,find_part_entry)(u, space, 0);
 	if (!entry)
@@ -206,7 +211,7 @@ error:
 /* Add "part" to "u", where "u" is assumed not to already have
  * a part that is defined on the same space as "part".
  */
-__isl_give UNION *FN(FN(UNION,add),PARTS)(__isl_take UNION *u,
+__isl_give UNION *FN(FN(UNION,add),BASE)(__isl_take UNION *u,
 	__isl_take PART *part)
 {
 	return FN(UNION,add_part_generic)(u, part, 1);
@@ -268,7 +273,7 @@ static isl_stat FN(UNION,transform_entry)(__isl_take PART *part, void *user)
 	S(UNION,transform_data) *data = (S(UNION,transform_data) *)user;
 
 	part = data->fn(part, data->user);
-	data->res = FN(FN(UNION,add),PARTS)(data->res, part);
+	data->res = FN(FN(UNION,add),BASE)(data->res, part);
 	if (!data->res)
 		return isl_stat_error;
 
@@ -285,7 +290,7 @@ static __isl_give UNION *FN(UNION,transform_space)(__isl_take UNION *u,
 	S(UNION,transform_data) data = { fn, user };
 
 	data.res = FN(UNION,alloc_same_size_on_space)(u, space);
-	if (FN(FN(UNION,foreach),PARTS)(u,
+	if (FN(FN(UNION,foreach),BASE)(u,
 					&FN(UNION,transform_entry), &data) < 0)
 		data.res = FN(UNION,free)(data.res);
 	FN(UNION,free)(u);
@@ -398,7 +403,7 @@ static __isl_give UNION *FN(UNION,realign_domain)(__isl_take UNION *u,
 	if (!u || !r)
 		goto error;
 
-	space = isl_space_copy(r->dim);
+	space = isl_reordering_get_space(r);
 	u = FN(UNION,transform_space)(u, space, &FN(UNION,align_entry), r);
 	isl_reordering_free(r);
 	return u;
@@ -413,17 +418,20 @@ error:
 __isl_give UNION *FN(UNION,align_params)(__isl_take UNION *u,
 	__isl_take isl_space *model)
 {
+	isl_bool equal_params;
 	isl_reordering *r;
 
 	if (!u || !model)
 		goto error;
 
-	if (isl_space_match(u->space, isl_dim_param, model, isl_dim_param)) {
+	equal_params = isl_space_has_equal_params(u->space, model);
+	if (equal_params < 0)
+		goto error;
+	if (equal_params) {
 		isl_space_free(model);
 		return u;
 	}
 
-	model = isl_space_params(model);
 	r = isl_parameter_alignment_reordering(u->space, model);
 	isl_space_free(model);
 
@@ -467,7 +475,7 @@ static __isl_give UNION *FN(UNION,union_add_)(__isl_take UNION *u1,
 	if (!u1 || !u2)
 		goto error;
 
-	if (FN(FN(UNION,foreach),PARTS)(u2, &FN(UNION,union_add_part), &u1) < 0)
+	if (FN(FN(UNION,foreach),BASE)(u2, &FN(UNION,union_add_part), &u1) < 0)
 		goto error;
 
 	FN(UNION,free)(u2);
@@ -479,23 +487,25 @@ error:
 	return NULL;
 }
 
-__isl_give UNION *FN(FN(UNION,from),PARTS)(__isl_take PART *part)
+__isl_give UNION *FN(FN(UNION,from),BASE)(__isl_take PART *part)
 {
-	isl_space *dim;
+	isl_space *space;
 	UNION *u;
 
 	if (!part)
 		return NULL;
 
-	dim = FN(PART,get_space)(part);
-	dim = isl_space_drop_dims(dim, isl_dim_in, 0, isl_space_dim(dim, isl_dim_in));
-	dim = isl_space_drop_dims(dim, isl_dim_out, 0, isl_space_dim(dim, isl_dim_out));
+	space = FN(PART,get_space)(part);
+	space = isl_space_drop_dims(space, isl_dim_in, 0,
+					isl_space_dim(space, isl_dim_in));
+	space = isl_space_drop_dims(space, isl_dim_out, 0,
+					isl_space_dim(space, isl_dim_out));
 #ifdef HAS_TYPE
-	u = FN(UNION,ZERO)(dim, part->type);
+	u = FN(UNION,ZERO)(space, part->type);
 #else
-	u = FN(UNION,ZERO)(dim);
+	u = FN(UNION,ZERO)(space);
 #endif
-	u = FN(FN(UNION,add),PARTS)(u, part);
+	u = FN(FN(UNION,add),BASE)(u, part);
 
 	return u;
 }
@@ -536,7 +546,7 @@ static isl_stat FN(UNION,match_bin_entry)(__isl_take PART *part, void *user)
 
 	part = data->fn(part, FN(PART, copy)(entry2->data));
 
-	data->res = FN(FN(UNION,add),PARTS)(data->res, part);
+	data->res = FN(FN(UNION,add),BASE)(data->res, part);
 	if (!data->res)
 		return isl_stat_error;
 
@@ -570,7 +580,7 @@ static __isl_give UNION *FN(UNION,match_bin_op)(__isl_take UNION *u1,
 
 	data.u2 = u2;
 	data.res = FN(UNION,alloc_same_size)(u1);
-	if (FN(FN(UNION,foreach),PARTS)(u1,
+	if (FN(FN(UNION,foreach),BASE)(u1,
 				    &FN(UNION,match_bin_entry), &data) < 0)
 		goto error;
 
@@ -668,12 +678,12 @@ S(UNION,match_domain_data) {
 	__isl_give PW *(*fn)(__isl_take PW*, __isl_take isl_set*);
 };
 
-static int FN(UNION,set_has_dim)(const void *entry, const void *val)
+static int FN(UNION,set_has_space)(const void *entry, const void *val)
 {
 	isl_set *set = (isl_set *)entry;
-	isl_space *dim = (isl_space *)val;
+	isl_space *space = (isl_space *)val;
 
-	return isl_space_is_equal(set->dim, dim);
+	return isl_space_is_equal(set->dim, space);
 }
 
 /* Find the set in data->uset that lives in the same space as the domain
@@ -690,7 +700,7 @@ static isl_stat FN(UNION,match_domain_entry)(__isl_take PART *part, void *user)
 	space = FN(PART,get_domain_space)(part);
 	hash = isl_space_get_hash(space);
 	entry2 = isl_hash_table_find(data->uset->dim->ctx, &data->uset->table,
-				     hash, &FN(UNION,set_has_dim), space, 0);
+				     hash, &FN(UNION,set_has_space), space, 0);
 	isl_space_free(space);
 	if (!entry2) {
 		FN(PART,free)(part);
@@ -699,7 +709,7 @@ static isl_stat FN(UNION,match_domain_entry)(__isl_take PART *part, void *user)
 
 	part = data->fn(part, isl_set_copy(entry2->data));
 
-	data->res = FN(FN(UNION,add),PARTS)(data->res, part);
+	data->res = FN(FN(UNION,add),BASE)(data->res, part);
 	if (!data->res)
 		return isl_stat_error;
 
@@ -724,7 +734,7 @@ static __isl_give UNION *FN(UNION,match_domain_op)(__isl_take UNION *u,
 
 	data.uset = uset;
 	data.res = FN(UNION,alloc_same_size)(u);
-	if (FN(FN(UNION,foreach),PARTS)(u,
+	if (FN(FN(UNION,foreach),BASE)(u,
 				   &FN(UNION,match_domain_entry), &data) < 0)
 		goto error;
 
@@ -829,7 +839,7 @@ __isl_give isl_union_set *FN(UNION,domain)(__isl_take UNION *u)
 	isl_union_set *uset;
 
 	uset = isl_union_set_empty(FN(UNION,get_space)(u));
-	if (FN(FN(UNION,foreach),PARTS)(u, &FN(UNION,domain_entry), &uset) < 0)
+	if (FN(FN(UNION,foreach),BASE)(u, &FN(UNION,domain_entry), &uset) < 0)
 		goto error;
 
 	FN(UNION,free)(u);
@@ -862,38 +872,6 @@ static __isl_give UNION *FN(UNION,negate_type)(__isl_take UNION *u)
 }
 #endif
 
-static __isl_give PART *FN(UNION,mul_isl_int_entry)(__isl_take PART *part,
-	void *user)
-{
-	isl_int *v = user;
-
-	return FN(PW,mul_isl_int)(part, *v);
-}
-
-__isl_give UNION *FN(UNION,mul_isl_int)(__isl_take UNION *u, isl_int v)
-{
-	if (isl_int_is_one(v))
-		return u;
-
-	if (DEFAULT_IS_ZERO && u && isl_int_is_zero(v)) {
-		UNION *zero;
-		isl_space *dim = FN(UNION,get_space)(u);
-#ifdef HAS_TYPE
-		zero = FN(UNION,ZERO)(dim, u->type);
-#else
-		zero = FN(UNION,ZERO)(dim);
-#endif
-		FN(UNION,free)(u);
-		return zero;
-	}
-
-	u = FN(UNION,transform_inplace)(u, &FN(UNION,mul_isl_int_entry), &v);
-	if (isl_int_is_neg(v))
-		u = FN(UNION,negate_type)(u);
-
-	return u;
-}
-
 /* Multiply "part" by the isl_val "user" and return the result.
  */
 static __isl_give PART *FN(UNION,scale_val_entry)(__isl_take PART *part,
@@ -1017,7 +995,7 @@ static isl_stat FN(UNION,plain_is_equal_entry)(void **entry, void *user)
 isl_bool FN(UNION,plain_is_equal)(__isl_keep UNION *u1, __isl_keep UNION *u2)
 {
 	S(UNION,plain_is_equal_data) data = { NULL, isl_bool_true };
-	int n1, n2;
+	isl_size n1, n2;
 
 	if (!u1 || !u2)
 		return isl_bool_error;
@@ -1025,8 +1003,8 @@ isl_bool FN(UNION,plain_is_equal)(__isl_keep UNION *u1, __isl_keep UNION *u2)
 		return isl_bool_true;
 	if (u1->table.n != u2->table.n)
 		return isl_bool_false;
-	n1 = FN(FN(UNION,n),PARTS)(u1);
-	n2 = FN(FN(UNION,n),PARTS)(u2);
+	n1 = FN(FN(UNION,n),BASE)(u1);
+	n2 = FN(FN(UNION,n),BASE)(u2);
 	if (n1 < 0 || n2 < 0)
 		return isl_bool_error;
 	if (n1 != n2)
@@ -1055,6 +1033,39 @@ error:
 	return isl_bool_error;
 }
 
+/* Check whether the element that "entry" points to involves any NaNs and
+ * store the result in *nan.
+ * Abort as soon as one such element has been found.
+ */
+static isl_stat FN(UNION,involves_nan_entry)(void **entry, void *user)
+{
+	isl_bool *nan = user;
+	PW *pw = *entry;
+
+	*nan = FN(PW,involves_nan)(pw);
+	if (*nan < 0 || !nan)
+		return isl_stat_error;
+
+	return isl_stat_ok;
+}
+
+/* Does "u" involve any NaNs?
+ */
+isl_bool FN(UNION,involves_nan)(__isl_keep UNION *u)
+{
+	isl_bool nan = isl_bool_false;
+
+	if (!u)
+		return isl_bool_error;
+
+	if (FN(UNION,foreach_inplace)(u,
+				    &FN(UNION,involves_nan_entry), &nan) < 0 &&
+	    !nan)
+		return isl_bool_error;
+
+	return nan;
+}
+
 /* Internal data structure for isl_union_*_drop_dims.
  * type, first and n are passed to isl_*_drop_dims.
  */
@@ -1161,3 +1172,39 @@ __isl_give UNION *FN(UNION,reset_user)(__isl_take UNION *u)
 	return FN(UNION,transform_space)(u, space, &FN(UNION,reset_user_entry),
 					NULL);
 }
+
+/* Add the base expression held by "entry" to "list".
+ */
+static isl_stat FN(UNION,add_to_list)(void **entry, void *user)
+{
+	PW *pw = *entry;
+	LIST(PART) **list = user;
+
+	*list = FN(LIST(PART),add)(*list, FN(PART,copy)(pw));
+	if (!*list)
+		return isl_stat_error;
+
+	return isl_stat_ok;
+}
+
+/* Return a list containing all the base expressions in "u".
+ *
+ * First construct a list of the appropriate size and
+ * then add all the elements.
+ */
+__isl_give LIST(PART) *FN(FN(UNION,get),LIST(BASE))(__isl_keep UNION *u)
+{
+	isl_size n;
+	LIST(PART) *list;
+
+	if (!u)
+		return NULL;
+	n = FN(FN(UNION,n),BASE)(u);
+	if (n < 0)
+		return NULL;
+	list = FN(LIST(PART),alloc)(FN(UNION,get_ctx(u)), n);
+	if (FN(UNION,foreach_inplace)(u, &FN(UNION,add_to_list), &list) < 0)
+		return FN(LIST(PART),free)(list);
+
+	return list;
+}
